/*
htop
(C) 2004 Hisham H. Muhammad
Released under the GNU GPL, see the COPYING file
in the source distribution for its full text.
*/

#include "ListBox.h"
#include "StringList.h"

#include <math.h>

#ifndef MIN
#define MIN(a,b) ((a)<(b)?(a):(b))
#endif

#include <curses.h>
//#link curses

/*{
typedef struct ListBox_ {
   int x, y, w, h;
   WINDOW* window;
   StringList* items;
   int selected;
   int scrollV, scrollH;
   int oldSelected;
   bool needsRedraw;
} ListBox;
}*/

ListBox* ListBox_new(int x, int y, int w, int h) {
   ListBox* this;
   this = malloc(sizeof(ListBox));
   this->x = x;
   this->y = y;
   this->w = w;
   this->h = h;
   this->window = subwin(stdscr, h, w, y, x);
   this->items = StringList_new();
   this->scrollV = 0;
   this->scrollH = 0;
   this->selected = 0;
   this->oldSelected = 0;
   this->needsRedraw = TRUE;
   return this;
}

void ListBox_delete(ListBox* this) {
   StringList_delete(this->items);
   delwin(this->window);
   free(this);
}

void ListBox_resize(ListBox* this, int w, int h) {
   this->w = w;
   this->h = h;
   wresize(this->window, h, w);
}

void ListBox_prune(ListBox* this) {
   StringList_prune(this->items);
   this->scrollV = 0;
   this->selected = 0;
   this->oldSelected = 0;
   this->needsRedraw = TRUE;
}

void ListBox_add(ListBox* this, char* entry) {
   StringList_add(this->items, entry);
}

void ListBox_set(ListBox* this, int i, char* entry) {
   StringList_set(this->items, i, entry);
}

int ListBox_getSelected(ListBox* this) {
   return this->selected;
}

void ListBox_setSelected(ListBox* this, int selected) {
   selected = MIN(StringList_size(this->items), selected);
   this->selected = selected;
}

void ListBox_draw(ListBox* this) {
   int first, last;
   int itemCount = StringList_size(this->items);
   int scrollH = this->scrollH;
   first = this->scrollV;
   if (this->h > itemCount) {
      last = this->scrollV + itemCount;
      wmove(this->window, last, 0);
      wclrtobot(this->window);
   } else {
      last = this->scrollV + this->h;
   }
   if (this->selected < first) {
      first = this->selected;
      this->scrollV = first;
      this->needsRedraw = TRUE;
   }
   if (this->selected >= last) {
      last = this->selected + 1;
      first = last - this->h;
      this->scrollV = first;
      this->needsRedraw = TRUE;
   }
   if (this->needsRedraw) {
      for(int i = first, j = 0; i < last; i++, j++) {
         char* itemRef = StringList_getRef(this->items, i);
         if (scrollH < strlen(itemRef)) {
            if (i == this->selected) {
               wattron(this->window, A_REVERSE);
               mvwhline(this->window, j, 0, ' ', this->w);
               mvwaddnstr(this->window, j, 0, itemRef + scrollH, this->w);
               wattroff(this->window, A_REVERSE);
            } else {
               mvwhline(this->window, j, 0, ' ', this->w);
               mvwaddnstr(this->window, j, 0, itemRef + scrollH, this->w);
            }
         }
      }
      this->needsRedraw = FALSE;
   } else {
      char* oldRef = StringList_getRef(this->items, this->oldSelected);
      char* newRef = StringList_getRef(this->items, this->selected);
      if (scrollH < strlen(oldRef)) {
         mvwhline(this->window, this->oldSelected - this->scrollV, 0, ' ', this->w);
         mvwaddnstr(this->window, this->oldSelected - this->scrollV, 0, oldRef + this->scrollH, this->w);
      }
      if (scrollH < strlen(newRef)) {
         wattron(this->window, A_REVERSE);
         mvwhline(this->window, this->selected - this->scrollV, 0, ' ', this->w);
         mvwaddnstr(this->window, this->selected - this->scrollV, 0, newRef + this->scrollH, this->w);
         wattroff(this->window, A_REVERSE);
      }
   }
   this->oldSelected = this->selected;
   move(0, 0);
   wrefresh(this->window);
}

void ListBox_onKey(ListBox* this, int key) {
   switch (key) {
   case KEY_DOWN:
      if (this->selected + 1 < StringList_size(this->items))
         this->selected++;
      break;
   case KEY_UP:
      if (this->selected > 0)
         this->selected--;
      break;
   case KEY_LEFT:
      if (this->scrollH > 0) {
         this->scrollH -= 10;
         this->needsRedraw = TRUE;
      }
      break;
   case KEY_RIGHT:
      this->scrollH += 10;
      this->needsRedraw = TRUE;
      break;
   case KEY_PPAGE:
      this->selected -= this->h;
      if (this->selected < 0)
         this->selected = 0;
      break;
   case KEY_NPAGE:
      this->selected += this->h;
      int size = StringList_size(this->items);
      if (this->selected >= size)
         this->selected = size - 1;
      break;
   case KEY_HOME:
      this->selected = 0;
      break;
   case KEY_END:
      this->selected = StringList_size(this->items) - 1;
      break;
   }
}
