/*
htop
(C) 2004 Hisham H. Muhammad
Released under the GNU GPL, see the COPYING file
in the source distribution for its full text.
*/

#include "ProcessList.h"
#include "Process.h"
#include "ProcessFilter.h"
#include "Vector.h"
#include "UsersTable.h"
#include "Hashtable.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

/*{
#ifndef PROCDIR
#define PROCDIR "/proc"
#endif

#ifndef PROCSTATFILE
#define PROCSTATFILE "/proc/stat"
#endif

#ifndef PROCMEMINFOFILE
#define PROCMEMINFOFILE "/proc/meminfo"
#endif
}*/

/*{
typedef struct ProcessList_ {
   Vector* processes;
   Hashtable* cmdlineCache;
   ProcessFilter* filter;
   Process* prototype;
   long int totalTime;
   long int userTime;
   long int systemTime;
   long int idleTime;
   long int niceTime;
   long int totalPeriod;
   long int userPeriod;
   long int systemPeriod;
   long int idlePeriod;
   long int nicePeriod;
   long int totalMem;
   long int usedMem;
   long int freeMem;
   long int sharedMem;
   long int buffersMem;
   long int cachedMem;
   long int totalSwap;
   long int usedSwap;
   long int freeSwap;
} ProcessList;
}*/

ProcessList* ProcessList_new(ProcessFilter* filter) {
   ProcessList* this;
   this = malloc(sizeof(ProcessList));
   this->processes = Vector_new();
   this->cmdlineCache = Hashtable_new(20);
   this->filter = filter;
   this->prototype = Process_new(this->filter);
   return this;
}

void ProcessList_delete(ProcessList* this) {
   Hashtable_delete(this->cmdlineCache);
   Vector_delete(this->processes);
   free(this);
}

void ProcessList_prune(ProcessList* this) {
   Vector_prune(this->processes);
}

void ProcessList_add(ProcessList* this, Process* p) {
   Vector_add(this->processes, p);
}

Process* ProcessList_get(ProcessList* this, int index) {
   return (Process*) (Vector_get(this->processes, index));
}

int ProcessList_size(ProcessList* this) {
   return (Vector_size(this->processes));
}

void ProcessList_scan(ProcessList* this, UsersTable* ut) {
   DIR* proc;
   struct dirent* entry;
   Process* prototype = this->prototype;
   int index = 0;
   long int usertime, nicetime, systemtime, idletime, totaltime;
   long int totalperiod;

   FILE* status;
   char buffer[128];
   status = fopen(PROCMEMINFOFILE, "r");
   assert(status != NULL);
   while (!feof(status)) {
      fgets(buffer, 128, status);
      if (String_startsWith(buffer, "MemTotal:")) {
         sscanf(buffer, "MemTotal: %ld kB", &this->totalMem);
      } else if (String_startsWith(buffer, "MemFree:")) {
         sscanf(buffer, "MemFree: %ld kB", &this->freeMem);
         this->usedMem = this->totalMem - this->freeMem;
      } else if (String_startsWith(buffer, "MemShared:")) {
         sscanf(buffer, "MemShared: %ld kB", &this->sharedMem);
      } else if (String_startsWith(buffer, "Buffers:")) {
         sscanf(buffer, "Buffers: %ld kB", &this->buffersMem);
      } else if (String_startsWith(buffer, "Cached:")) {
         sscanf(buffer, "Cached: %ld kB", &this->cachedMem);
      } else if (String_startsWith(buffer, "SwapTotal:")) {
         sscanf(buffer, "SwapTotal: %ld kB", &this->totalSwap);
      } else if (String_startsWith(buffer, "SwapFree:")) {
         long int swapFree;
         sscanf(buffer, "SwapFree: %ld kB", &swapFree);
         this->usedSwap = this->totalSwap - swapFree;
      }
   }
   fclose(status);

   status = fopen(PROCSTATFILE, "r");
   assert(status != NULL);
   fscanf(status, "cpu  %ld %ld %ld %ld", &usertime, &nicetime, &systemtime, &idletime);
   totaltime = usertime + nicetime + systemtime + idletime;
   totalperiod = totaltime - this->totalTime;
   fclose(status);

   proc = opendir(PROCDIR);
   assert(proc != NULL);
   while ((entry = readdir(proc)) != NULL) {
      char* name = entry->d_name;
      int pid;
      // filename is a number: process directory
      pid = atoi(name);
      if (pid > 0) {
         FILE* status;
         const int MAX_NAME = 128;
         char statusfilename[MAX_NAME+1];
         long int zero;

         Process* process;
         checkPidInList: {
            Process* indexProcess = (Process*) Vector_get(this->processes, index);
            if (!indexProcess) {
               process = Process_clone(prototype);
               Vector_add(this->processes, process);
            } else {
               int ipid = indexProcess->pid;
               if (pid == ipid) {
                  process = indexProcess;
               } else if (pid < ipid) {
                  process = Process_clone(prototype);
                  Vector_insert(this->processes, index, process);
               } else { // pid > ipid
                  process = Process_clone(prototype);
                  Process* removed = (Process*) Vector_remove(this->processes, index);
                  if (removed) {
                     free(removed);
                     // TODO: the hash seems not to be working.
                     //Hashtable_remove(this->cmdlineCache, ipid);
                     goto checkPidInList;
                  }
               }
            }
         }
         index++;
         struct stat sstat;
         snprintf(statusfilename, MAX_NAME, "%s/%s/stat", PROCDIR, name);
         stat(statusfilename, &sstat);
         char* username = UsersTable_getRef(ut, sstat.st_uid);
         if (username) {
            strncpy(process->user, username, PROCESS_USER_LEN);
         } else {
            // TODO: Handle NIS users
            snprintf(process->user, PROCESS_USER_LEN, "%d", sstat.st_uid);
         }

         int lasttimes = (process->utime + process->stime);

         status = fopen(statusfilename, "r");
         assert(status != NULL);
         // TODO: handle process names with spaces.
         fscanf(status, "%d %s %c %d %d %d "
                        "%d %d %lu %lu %lu %lu %lu "
                        "%lu %lu %ld %ld %ld %ld "
                        "%ld %ld %lu %lu %ld %lu "
                        "%lu %lu %lu %lu %lu %lu "
                        "%lu %lu %lu %lu %lu %lu "
                        "%d %d",
                &process->pid, (char*)&process->comm, &process->state, &process->ppid, &process->pgrp, &process->session,
                &process->tty_nr, &process->tpgid, &process->flags, &process->minflt, &process->cminflt, &process->majflt, &process->cmajflt,
                &process->utime, &process->stime, &process->cutime, &process->cstime, &process->priority, &process->nice,
                &zero, &process->itrealvalue, &process->starttime, &process->vsize, &process->rss, &process->rlim,
                &process->startcode, &process->endcode, &process->startstack, &process->kstkesp, &process->kstkeip, &process->signal,
                &process->blocked, &process->sigignore, &process->sigcatch, &process->wchan, &process->nswap, &process->cnswap,
                &process->exit_signal, &process->processor);
         // TODO: Check stat fields on kernel 2.6
         // assert(zero == 0);
         fclose(status);

         process->percent_cpu = (process->utime + process->stime - lasttimes) / (float)totalperiod * 100.0;

         // TODO: the hash seems not to be working.
         /*
         char* cmdline = Hashtable_get(this->cmdlineCache, process->pid);
         if (cmdline) {
            strcpy(process->comm, cmdline);
         } else {
         */
            snprintf(statusfilename, MAX_NAME, "%s/%s/cmdline", PROCDIR, name);
            status = fopen(statusfilename, "r");
            assert(status != NULL);
            char* cmdline = (char*) &(process->comm);
            int amtRead = fread(cmdline, 1, PROCESS_COMM_LEN - 1, status);
            if (amtRead > 0) {
               for (int i = 0; i < PROCESS_COMM_LEN - 1; i++)
                  if (cmdline[i] == '\0')
                     cmdline[i] = ' ';
               cmdline[amtRead] = '\0';
            }
            fclose(status);
         // TODO: the hash seems not to be working.
         /*
            Hashtable_put(this->cmdlineCache, process->pid, String_copy(process->comm));
         }
         */

         /*
         // Just format the name gotten from stat.
         int commLen = strlen(process->comm) - 2;
         memmove(process->comm, process->comm + 1, commLen);
         process->comm[commLen] = 0;
         */

         snprintf(statusfilename, MAX_NAME, "%s/%s/statm", PROCDIR, name);
         status = fopen(statusfilename, "r");
         assert(status != NULL);
         fscanf(status, "%d %d %d %d %d %d %d",
                &process->m_size, &process->m_resident, &process->m_share, 
                &process->m_trs, &process->m_drs, &process->m_lrs, 
                &process->m_dt);
         fclose(status);
      }
   }
   closedir(proc);

   Process* indexProcess;
   do {
      indexProcess = (Process*) Vector_get(this->processes, index);
      if (indexProcess) {
         Process* removed = (Process*) Vector_remove(this->processes, index);
         if (removed) {
            free(removed);
            // TODO: the hash seems not to be working.
            /*
            int ipid = indexProcess->pid;
            Hashtable_remove(this->cmdlineCache, ipid);
            */
         }
      }
   } while (indexProcess);

   totalperiod = totaltime - this->totalTime;
   long int userperiod = usertime - this->userTime;
   long int niceperiod = nicetime - this->niceTime;
   long int systemperiod = systemtime - this->systemTime;
   long int idleperiod = idletime - this->idleTime;

   this->totalTime = totaltime;
   this->userTime = usertime;
   this->niceTime = nicetime;
   this->systemTime = systemtime;
   this->idleTime = idletime;

   this->totalPeriod = totalperiod;
   this->userPeriod = userperiod;
   this->nicePeriod = niceperiod;
   this->systemPeriod = systemperiod;
   this->idlePeriod = idleperiod;
}
