/*
htop
(C) 2004 Hisham H. Muhammad
Released under the GNU GPL, see the COPYING file
in the source distribution for its full text.
*/

#include "ProcessList.h"
#include "CRT.h"
#include "ListBox.h"
#include "UsersTable.h"

#include <unistd.h>
#include <math.h>

//#link m

#define NICE_CPU_COLOR BLUE_PAIR
#define USER_CPU_COLOR GREEN_PAIR
#define SYSTEM_CPU_COLOR RED_PAIR

#define USED_MEMORY_COLOR GREEN_PAIR
#define BUFFERS_MEMORY_COLOR BLUE_PAIR
#define CACHED_MEMORY_COLOR BROWN_PAIR

#define USED_SWAP_COLOR RED_PAIR

#define HEADER_COLOR INV_PAIR

#ifndef MIN
#define MIN(a,b) ((a)<(b)?(a):(b))
#endif
#ifndef MAX
#define MAX(a,b) ((a)>(b)?(a):(b))
#endif

// TODO: cleanup interface
void statusMessage(char* msg) {
   mvaddstr(0, 4, msg);
   refresh();
}

void drawBar(int x, int y, int w, char* displayValue, int items, ...) {
   assert(displayValue != NULL);
   attron(A_BOLD);
   mvaddch(y, x, '[');
   mvaddch(y, x + w, ']');
   attroff(A_BOLD);
   w--;
   x++;
   char bar[w];
   int attrs[10];
   int blockSizes[10];
   for (int i = 0; i < w; i++)
      bar[i] = ' ';
   sprintf(bar + (w-strlen(displayValue)), "%s", displayValue);
   va_list ap;
   va_start(ap, items);
   double total = 0.0;
   int offset = 0;
   for (int i = 0; i < items; i++) {
      attrs[i] = va_arg(ap, int);
      double value = va_arg(ap, double);
      blockSizes[i] = ceil(value * w);
      int nextOffset = offset + blockSizes[i];
      // Control against invalid values
      nextOffset = MAX(nextOffset, 0);
      nextOffset = MIN(nextOffset, w);
      for (int j = offset; j < nextOffset; j++)
         if (bar[j] == ' ')
            bar[j] = '|';
      offset = nextOffset;
      total += value;
   }
   va_end(ap);
   offset = 0;
   for (int i = 0; i < items; i++) {
      attron(attrs[i]);
      mvaddnstr(y, x + offset, bar + offset, blockSizes[i]);
      attroff(attrs[i]);
      offset += blockSizes[i];
   }
   attron(A_BOLD);
   if (offset < w) {
      attron(COLOR_PAIR(BLACK_PAIR));
      mvaddnstr(y, x + offset, bar + offset, w - offset);
      attroff(COLOR_PAIR(BLACK_PAIR));
   }
   attroff(A_BOLD);
}

void drawCPUBar(ProcessList* pl, int x, int y, int w) {
   double niceCPU = pl->nicePeriod / (double)pl->totalPeriod;
   double userCPU = pl->userPeriod / (double)pl->totalPeriod;
   double systemCPU = pl->systemPeriod / (double)pl->totalPeriod;
   char total[8];
   snprintf(total, 7, "%5.1f%%", (niceCPU+userCPU+systemCPU)*100.0 );
   drawBar(x+3, y, w, total, 3,
           COLOR_PAIR(NICE_CPU_COLOR), niceCPU,
           COLOR_PAIR(USER_CPU_COLOR), userCPU,
           COLOR_PAIR(SYSTEM_CPU_COLOR), systemCPU);
   mvaddstr(y, x, "CPU");
}

void drawMemoryBar(ProcessList* pl, int x, int y, int w) {
   double totalMem = (double)pl->totalMem;
   long int usedMem = pl->usedMem, buffersMem = pl->buffersMem, cachedMem = pl->cachedMem;
   usedMem -= buffersMem + cachedMem;
   char total[13];
   snprintf(total, 12, "%9ldKB", usedMem);
   drawBar(x+3, y, w, total, 3,
           COLOR_PAIR(USED_MEMORY_COLOR), usedMem / totalMem,
           COLOR_PAIR(BUFFERS_MEMORY_COLOR), buffersMem / totalMem,
           COLOR_PAIR(CACHED_MEMORY_COLOR), cachedMem / totalMem);
   mvaddstr(y, x, "Mem");
}

void drawSwapBar(ProcessList* pl, int x, int y, int w) {
   double totalSwap = (double)pl->totalSwap;
   long int usedSwap = pl->usedSwap;
   char total[13];
   snprintf(total, 12, "%9ldKB", usedSwap / 1024);
   drawBar(x+3, y, w, total, 1,
           COLOR_PAIR(USED_SWAP_COLOR), usedSwap / totalSwap);
   mvaddstr(y, x, "Swp");
}

void drawHeader(char* header, int x, int y, int w, int scrollH) {
   attron(COLOR_PAIR(INV_PAIR));
   move(y, x);
   hline(' ', w);
   if (scrollH < strlen(header)) {
      mvaddnstr(y, x, header + scrollH, w);
   }
   attroff(COLOR_PAIR(INV_PAIR));
}

int main() {
   ListBox* lb;
   int quit = 0;
   int refreshTimeout = 0;
   int resetRefreshTimeout = 1;
   ProcessFilter* pf;

   char incSearchBuffer[20];
   int incSearchIndex = 0;
   incSearchBuffer[0] = 0;
   bool incSearchMode = FALSE;

   ProcessList* pl = NULL;
   UsersTable* ut = UsersTable_new();
   UsersTable_scan(ut);
   CRT_init();
   lb = ListBox_new(0, 6, COLS, LINES - 6);
   pf = ProcessFilter_new();
   pl = ProcessList_new(pf);
   ProcessList_scan(pl, ut);
   char* header = ((Object*)pf)->display((Object*)pf);
  
   while (!quit) {
      if (refreshTimeout == 0) {
         incSearchIndex = 0;
         incSearchBuffer[0] = 0;
         int currPos = ListBox_getSelected(lb);
         int currPid = 0;
         int currScrollV = lb->scrollV;
         currPid = ProcessList_get(pl, currPos)->pid;
         ProcessList_scan(pl, ut);
         ListBox_prune(lb);
         int size = ProcessList_size(pl);
         for (int i = 0; i < size; i++) {
            Process* p = ProcessList_get(pl, i);
            ListBox_set(lb, i, ((Object*)p)->display((Object*)p));
            if (p->pid == currPid) {
               ListBox_setSelected(lb, i);
               lb->scrollV = currScrollV;
            }
         }
         refreshTimeout = resetRefreshTimeout;
      }
      
      drawCPUBar(pl, 2, 1, COLS/2 - 4);
      drawMemoryBar(pl, 2, 2, COLS/2 - 4);
      drawSwapBar(pl, 2, 3, COLS/2 - 4);
      drawHeader(header, 0, 5, COLS, lb->scrollH);

      ListBox_draw(lb);
      int ch = getch();
      if (incSearchMode) {
         if (ch >= 'a' && ch <= 'z') {
            if (incSearchIndex == 19)
               break;
            incSearchBuffer[incSearchIndex] = ch;
            incSearchIndex++;
            incSearchBuffer[incSearchIndex] = 0;
            mvaddstr(0,1,incSearchBuffer);
            for (int i = 0; i < ProcessList_size(pl); i++) {
               Process* p = ProcessList_get(pl, i);
               if (String_startsWith_i(p->comm, incSearchBuffer)) {
                  ListBox_setSelected(lb, i);
                  break;
               }
            }
         } else if (ch == KEY_BACKSPACE) {
            if (incSearchIndex > 0) {
               incSearchIndex--;
               mvaddch(0, incSearchIndex+1, ' ');
               incSearchBuffer[incSearchIndex] = 0;
            }
          } else if (ch != ERR) {
            incSearchMode = FALSE;
            incSearchIndex = 0;
            incSearchBuffer[0] = 0;
            mvhline(0, 0, ' ', 21);
         }
      } else switch (ch) {
      case ERR:
         refreshTimeout--;
         break;
      case KEY_RESIZE:
         ListBox_resize(lb, COLS, LINES-5);
         break;
      case 'k':
      {
         statusMessage("Press enter to kill");
         int confirm = CRT_readKey();
         // TODO: handle other signals.
         int signal = 15;
         if (confirm == '9') {
            signal = 9;
            confirm = 13;
         }
         if (confirm == 13) {
            statusMessage("Sending signal...     ");
            int i = ListBox_getSelected(lb);
            Process* p = ProcessList_get(pl, i);
            Process_sendSignal(p, signal);
            ListBox_setSelected(lb, i);
            sleep(1);
            statusMessage("                      ");
            refreshTimeout = 0;
         } else {
            statusMessage("                      ");
         }
         break;
      }
      case 'q':
         quit = 1;
         break;
      case '[':
      {
         int i = ListBox_getSelected(lb);
         Process* p = ProcessList_get(pl, i);
         Process_setPriority(p, p->nice + 1);
         ListBox_set(lb, i, ((Object*)p)->display((Object*)p));
         break;
      }
      case ']':
      {
         int i = ListBox_getSelected(lb);
         Process* p = ProcessList_get(pl, i);
         Process_setPriority(p, p->nice - 1);
         ListBox_set(lb, i, ((Object*)p)->display((Object*)p));
         break;
      }
      case '/':
         mvaddch(0,0,'/');
         incSearchMode = TRUE;
         break;
      default:
         refreshTimeout = resetRefreshTimeout;
         ListBox_onKey(lb, ch);
         break;
      }
   }
   ListBox_delete(lb);
   CRT_done();
   return 0;
}
